@@ -24,6 +24,7 @@ from code42cli.options import sdk_options
 from code42cli.output_formats import DataFrameOutputFormatter
 from code42cli.output_formats import OutputFormat
 from code42cli.output_formats import OutputFormatter
+from code42cli.worker import create_worker_stats
 
 
 @click.group(cls=OrderedGroup)
@@ -83,9 +84,7 @@ def _deactivate_device(sdk, device_guid, change_device_name, purge_date):
     try:
         device = _change_device_activation(sdk, device_guid, "deactivate")
     except exceptions.Py42BadRequestError:
-        raise Code42CLIError(
-            "The device with GUID '{}' is in legal hold.".format(device_guid)
-        )
+        raise Code42CLIError(f"The device with GUID '{device_guid}' is in legal hold.")
     if purge_date:
         _update_cold_storage_purge_date(sdk, device_guid, purge_date)
     if change_device_name and not device.data["name"].startswith("deactivated_"):
@@ -113,12 +112,10 @@ def _change_device_activation(sdk, device_guid, cmd_str):
             sdk.devices.deactivate(device_id)
         return device
     except exceptions.Py42NotFoundError:
-        raise Code42CLIError(
-            "The device with GUID '{}' was not found.".format(device_guid)
-        )
+        raise Code42CLIError(f"The device with GUID '{device_guid}' was not found.")
     except exceptions.Py42ForbiddenError:
         raise Code42CLIError(
-            "Unable to {} the device with GUID '{}'.".format(cmd_str, device_guid)
+            f"Unable to {cmd_str} the device with GUID '{device_guid}'."
         )
 
 
@@ -337,6 +334,13 @@ def list_devices(
         org_uid,
         (include_backup_usage or include_total_storage),
     )
+    if exclude_most_recently_connected:
+        most_recent = (
+            df.sort_values(["userUid", "lastConnected"], ascending=False)
+            .groupby("userUid")
+            .head(exclude_most_recently_connected)
+        )
+        df = df.drop(most_recent.index)
     if last_connected_after:
         df = df.loc[to_datetime(df.lastConnected) > last_connected_after]
     if last_connected_before:
@@ -345,13 +349,6 @@ def list_devices(
         df = df.loc[to_datetime(df.creationDate) > created_after]
     if created_before:
         df = df.loc[to_datetime(df.creationDate) < created_before]
-    if exclude_most_recently_connected:
-        most_recent = (
-            df.sort_values(["userUid", "lastConnected"], ascending=False)
-            .groupby("userUid")
-            .head(exclude_most_recently_connected)
-        )
-        df = df.drop(most_recent.index)
     if include_total_storage:
         df = _add_storage_totals_to_dataframe(df, include_backup_usage)
     if include_settings:
@@ -509,14 +506,12 @@ def _add_backup_set_settings_to_dataframe(sdk, devices_dataframe):
     def handle_row(guid):
         try:
             current_device_settings = sdk.devices.get_settings(guid)
-        except Exception as e:
+        except Exception as err:
             return DataFrame.from_records(
                 [
                     {
                         "guid": guid,
-                        "ERROR": "Unable to retrieve device settings for {}: {}".format(
-                            guid, e
-                        ),
+                        "ERROR": f"Unable to retrieve device settings for {guid}: {err}",
                     }
                 ]
             )
@@ -578,9 +573,14 @@ bulk.add_command(devices_generate_template)
 @sdk_options()
 def bulk_deactivate(state, csv_rows, change_device_name, purge_date, format):
     """Deactivate all devices from the provided CSV containing a 'guid' column."""
+
+    # Initialize the SDK before starting any bulk processes
+    # to prevent multiple instances and having to enter 2fa multiple times.
     sdk = state.sdk
-    csv_rows[0]["deactivated"] = False
+
+    csv_rows[0]["deactivated"] = "False"
     formatter = OutputFormatter(format, {key: key for key in csv_rows[0].keys()})
+    stats = create_worker_stats(len(csv_rows))
     for row in csv_rows:
         row["change_device_name"] = change_device_name
         row["purge_date"] = purge_date
@@ -591,12 +591,17 @@ def bulk_deactivate(state, csv_rows, change_device_name, purge_date, format):
                 sdk, row["guid"], row["change_device_name"], row["purge_date"]
             )
             row["deactivated"] = "True"
-        except Exception as e:
-            row["deactivated"] = "False: {}".format(e)
+        except Exception as err:
+            row["deactivated"] = f"False: {err}"
+            stats.increment_total_errors()
         return row
 
     result_rows = run_bulk_process(
-        handle_row, csv_rows, progress_label="Deactivating devices:"
+        handle_row,
+        csv_rows,
+        progress_label="Deactivating devices:",
+        stats=stats,
+        raise_global_error=False,
     )
     formatter.echo_formatted_list(result_rows)
 
@@ -607,19 +612,29 @@ def bulk_deactivate(state, csv_rows, change_device_name, purge_date, format):
 @sdk_options()
 def bulk_reactivate(state, csv_rows, format):
     """Reactivate all devices from the provided CSV containing a 'guid' column."""
+
+    # Initialize the SDK before starting any bulk processes
+    # to prevent multiple instances and having to enter 2fa multiple times.
     sdk = state.sdk
-    csv_rows[0]["reactivated"] = False
+
+    csv_rows[0]["reactivated"] = "False"
     formatter = OutputFormatter(format, {key: key for key in csv_rows[0].keys()})
+    stats = create_worker_stats(len(csv_rows))
 
     def handle_row(**row):
         try:
             _reactivate_device(sdk, row["guid"])
             row["reactivated"] = "True"
-        except Exception as e:
-            row["reactivated"] = "False: {}".format(e)
+        except Exception as err:
+            row["reactivated"] = f"False: {err}"
+            stats.increment_total_errors()
         return row
 
     result_rows = run_bulk_process(
-        handle_row, csv_rows, progress_label="Reactivating devices:"
+        handle_row,
+        csv_rows,
+        progress_label="Reactivating devices:",
+        stats=stats,
+        raise_global_error=False,
     )
     formatter.echo_formatted_list(result_rows)
